package masterelection

import (
	"log"
	"time"

	"git.autistici.org/ale/radioai/third_party/github.com/coreos/go-etcd/etcd"
)

const (
	STATE_SLAVE = iota
	STATE_MASTER
)

func stateToString(state int) string {
	switch state {
	case STATE_SLAVE:
		return "slave"
	case STATE_MASTER:
		return "master"
	}
	return ""
}

type MasterElection struct {
	client *etcd.Client
	stop chan bool
	stopped bool

	Addr string
	Path string
	TTL  uint64

	State       int
	StateChange chan int
}

func NewMasterElection(client *etcd.Client, path, addr string, ttl uint64, sch chan int, stop chan bool) *MasterElection {
	if ttl < 2 {
		ttl = 2
	}
	return &MasterElection{
		client:      client,
		Path:        path,
		Addr:        addr,
		TTL:         ttl,
		State:       STATE_SLAVE,
		StateChange: sch,
		stop:        stop,
	}
}

func (m *MasterElection) IsMaster() bool {
	return m.State == STATE_MASTER
}

func (m *MasterElection) GetMasterAddr() string {
	responses, err := m.client.Get(m.Path)
	if err != nil || len(responses) != 1 {
		return ""
	}
	return responses[0].Value
}

func (m *MasterElection) setState(state int) {
	if m.State == state {
		return
	}
	log.Printf("masterelection: %s -> %s",
		stateToString(m.State),
		stateToString(state))
	if m.StateChange != nil {
		m.StateChange <- state
	}
	m.State = state
}

func (m *MasterElection) stopper() {
	<-m.stop

	// Tell the Run() goroutine to exit as soon as it can (we
	// could have simply used the 'stop' channel there but Watch()
	// ignores the stop channel if invoked without a receiving
	// channel and we'd have to refactor Run() to use a temp
	// channel for each invocation.
	m.stopped = true

	// Remove the lock file if we are the master.
	if m.State == STATE_MASTER {
		m.client.Delete(m.Path)
	}
}

func (m *MasterElection) Run() {
	go m.stopper()

	// Start as a slave.
	m.setState(STATE_SLAVE)

	halfttl := time.Second * time.Duration(m.TTL / 2)

	for !m.stopped {
		// Try to acquire the lock. If we are currently the
		// master, the previous value should be our own
		// address, otherwise it should be unset.
		prevValue := ""
		if m.State == STATE_MASTER {
			prevValue = m.Addr
		}
		if _, ok, _ := m.client.TestAndSet(m.Path, prevValue, m.Addr, m.TTL); ok {
			// Howdy, we're the master now. Wait a while
			// and renew our TTL.
			m.setState(STATE_MASTER)
			time.Sleep(halfttl)
		} else {
			// We're not the master. We could wait for a
			// DELETE but I'm not sure if that's what you
			// get on expiry, so we just wait for the
			// first event which will be another SET from
			// the current master. Oh well.
			m.setState(STATE_SLAVE)
			_, err := m.client.Watch(m.Path, 0, nil, nil)
			if err != nil {
				log.Printf("%s: watch error: %s", m.Path, err)
			}
		}
	}
}
